#include"cellmanagement.h"
#include"SWAP.h"
#include"IO.h"
#include"sched.h"
#include"string.h"
#define PI 3.1415926536

// Global variables
double* RVF;    // current positions, velocities, forces
double* Rlast;  // positions at last NL build
double* Rinit;  // initial positions (for calculation of MSD, etc.)
double* mass;   // particle masses
double* sigma;  // particle diameters
double* epsilon;    // particle energy/force prefactors
double* sqrtepsilon; // saves having to calculate sqrt(epsilon_i epsilon_j) every timestep
int* Z_array;   // array for counting Z
int* allneigh;  // neighbor list
short int* numneigh;    // number of neighbors/atom
int maxneigh;   // maximum number of neighbors/atom
short int* image;   // current particle image flags
short int* initimg; // initial image flags (for calculation of MSD, etc.)
double ALU[3];  // polydisperse-sigma-distribution parameters
int d;      // spatial dimension
int N;      // number of particles simulated
int ncell;  // number of cubic subcells along each dimension
int ncellprev;  // previous value of ncell (needed if cell volume V varies)
int totcells;   // total number of subcells
int nbs;    // nbs = 3 or 5 controls structure of cell linking
double skin;    // skin depth for NL builds
double cut, cutsq;  // for longer-range interactions
int neighlistbuilds;    // number of NL builds performed
double avgneigh, totavgneigh;   // current and overall average neighbors/particle
int nthreads;   // parallel simulations run on nthreads cores
const int nchunk = 100;     // Use schedule(static,nchunk) for integration of all EOM
double dt, dthalf;      // timestep, half-timestep for leapfrog, etc.
int nsteps;     // simulation runs for nsteps timesteps
int nbet;       // leapfrog this many steps between NL-build-checks
int nthermo;    // calculate thermodynamic data every nthermo timesteps
int nvhc;       // calculate van Hove corr. function every nvhc timesteps
int nbaro;      // apply barostat every nbaro timesteps
int nmin;       // energy-minimize every nmin timesteps
int nswap;       // SWAP every nswap timesteps
int nmsdstart;  // start MSD calculation after nmsdstart steps
int nrestart;   // write restart file every nrestart timesteps
int ndump;     // write coordinates every ndump timesteps
int step;   // current timestep
int lastswapstep;  // sometimes required for stability when SWAP employed
int rattlers, floaters; //calculating rattler and floating particles
long long int swapatt, swapsucc;    // number of attempted & successful swaps
double attfrac;  // attempt swaps for attfrac of particles
// every time swapmove() is called
double D;   // simulation cell side length
double Dinit;   // initial value of D (for calculation of MSD, etc.)
double lo, hi;  // positions of cell's lower, upper boundaries (typically -/+ D/2)
double rho, phi;    // particle number density and packing fraction
double totpartvol, vol;     // total particle volume & cell volume
double E, EP, EK;  // total, pair and kinetic energies
double T, P, Z, Znoratt;    // temperature, pressure, average coordination number
double Ttarg, Ptarg;    // target temperature and pressure for thermostat, barostat
double Tdot; // for runs with temperature ramp
double msd, namsd, ngp, maxdisp, f_ov;  // measures of diffusion
double min_actual_time; // total time elapsed in FIRE and total displacement for SD

// Functions defined in main.cpp
void leapfrog();    // Leapfrog integration with thermostat
void getforce(int i);   // Get force on atom i
void getthermo(bool outputflag);    // Calculate thermodynamic quantities
void getvanHove();  // calculate van Hove correlation function
void Berendsen(double Ptarg);   // Berendsen thermostat
inline double mypow3(double x); // Efficient exponentiation for generalized-LJ
void precompress(double Ptarg);

// Functions defined in other routines
extern double initRVFMSE(bool readresflag, const char *initresfname, const char *initRinitfname); // Initialize particle positions, diameters, velocities, image flags, etc.
extern double initmsig(double MVR); // initialize sigma/mass/epsilon distrib.
extern void dealloc();      // deallocate all large arrays
extern void FIRE(double Ethres, double fracDEcap);  // FIRE minimization
extern void steepest(double Ethres);    // steepest-descent minimization
extern void moddownhill(double Ethres, double fracDEcap);  // modified downhill minimization
extern void modminimize(double dtm);    // modified downhill minimization
extern void downhill(double Ethres);  // standard downhill minimization
extern void onestepdownhill(double dtm); // one-step-straight-downhill integration

// Class declarations
cell supercell;
swapMC SWAP;
IO readwrite;

// Parameters specific to this simulation
int n = 6; // generalized-LJ exponent
int alpha = 30;  // Morse exponent
const bool baroflag = false;    // True if applying barostat, otherwise false
const bool minflag = true;      // True if performing periodic E-minimizations
const bool vhcflag = false;     // True if outputting van Hove correlation function
const double tau_press = 10; // Berendsen barostat pressure damping time
const double tau_temp = 3.16228; // Berendsen thermostat temperature damping time
double dispcap = .025;      // Maximum single-step displacement cap for integration of various EOM


int main (int argc, const char *argv[]) {
    d = atoi(argv[1]);  // spatial dimension
    N = atoi(argv[2]);  // number of particles
    phi = atof(argv[3]);    // volume fraction
    Ttarg = atof(argv[4]);  // target temperature
    nsteps = atoi(argv[5]);     // # of steps in run
    nthreads = atoi(argv[6]);   // # of parallel threads
    attfrac = atof(argv[7]);   // fraction of atoms to attempt SWAP moves for / tau
// Can start simulation from a restart file and a file with Rinit values
// (latter so can continue calculation of msd, ngp, f_ov, etc.)
    const char *initresfname = argv[8];  // (optional) start simulation from this restart file
    const char *initRinitfname = argv[9]; // (optional) reference state for MSD calculation
    nmsdstart = atoi(argv[10]);  // start msd calculation after this many steps
    
    bool readresflag;
    if (!strcmp(initresfname,"null") && !strcmp(initRinitfname,"null")) readresflag = false;
    else readresflag = true;
    
// Initialize loop counters and various run-progress-related variables
    int i, j, k, l;
    long long int lastswapsucc;  // for checking whether need rebuild NL after swapping
    if (!readresflag) step = 0;   // MD timestep
    swapatt = 0;    swapsucc = 0;   // SWAP attempts and successes
// Timestep for generalized-LJ interactions
//   dt = 6.0/(125*n); //   for example, dt = .008 for n = 6
// Timestep for Morse interactions
    dt = 6.0/(125*alpha);
    dthalf = dt/2;
// Counters for NL-build tracking
    neighlistbuilds = 0;    totavgneigh = 0;
// Thermo/barostat and other simulation-control parameters: below outputs thermo data once/tau,
// calls barostat 25 times/tau, writes restart files once/(1000tau), etc.
    nthermo = static_cast<int>(1/dt);   // # of steps to run between calls to getthermo()
    nvhc = static_cast<int>(100/dt);     // # of steps to run between calls to getvanHove()
    nbaro = static_cast<int>(.04/dt);   // # of steps to run between calls to Berendsen()
    nrestart = static_cast<int>(1000/dt);   // # of steps between restart-file writes
    ndump = static_cast<int>(1000/dt);   // # of steps between coordinate dumps
    nbet = static_cast<int>(.008/dt);   // # of steps to run between NL-build checks
    nmin = static_cast<int>(10/dt);   // # of steps to run between energy minimizations
    nswap = static_cast<int>(1/dt);    // # of steps to run between SWAP MC passes
    
// Timers for high resolution timing
    chrono::duration<double> elapsed, totelapsed, nltime, totnltime, verltime, totverltime;
    chrono::duration<double> pushtime, mintime, totmintime, barotime, totbarotime;
    chrono::duration<double> swaptime, totswaptime;
    auto start = chrono::high_resolution_clock::now();
    
// Set maximum # of neighbors to twice the kissing number
    int kissnum[11] = {0,2,6,12,24,40,72,126,240,306,500};
    maxneigh = max(2*kissnum[d],200);   // kludge to reduce usage in large d
// Initalize all per-particle quantities (positions, diameters, velocities,
// image flags, neighbor arrays, etc.)
    totpartvol = initRVFMSE(readresflag,initresfname,initRinitfname);;
    int startstep = step;
    
// Set up neighboring: neighbor cutoff is at r_c(i,u) = cut*(max. particle diameter) + skin
    cut = 1.0;    cutsq = cut*cut;
 //   if (d < 7) skin = .25;      else skin = .33;
    skin = .25;
    ncell = floor(D/(cut*ALU[2] + skin));
    nbs = 3;     // Default: each subcell is liked to its 3^d - 1 neighbors
// Alternate: 5^d - 1 neighboring cells of half the size; use only for long-range attractive interactions
// THIS ALWAYS SLOWS CODE DOWN BC IT PRODUCES A ~ (10/3)^d increase in the effort to calculate neighindex
//    if (cut > 2 || d > 7) {  ncell = floor(D/((cut*ALU[2] + skin)/2));   nbs = 5;  }
    ncellprev = ncell;  // ncellprev gets used in NPT and/or phi-controlled MD
    cout << "N = " << N << " phi = " << phi << " D = " << D << " ncell = " << ncell << endl;
    
// Set up linked subcells
    supercell.setupsupercell();
// Set up SWAP Monte Carlo if needed
    if (attfrac) SWAP.setupswap();
// Setup file I/O
    readwrite.openfiles();

// Thread and start the simulation.  The master thread controls things but needsrebuild(),
// buildneighborlist(), leapfrog(), getthermo(), Berendsen(), etc. are all parallelized
    omp_set_num_threads(nthreads);
#pragma omp master
{
// Initialize neighbor lists
    cout << "Initializing neighbor lists." << endl;
    auto thisnlstart = chrono::high_resolution_clock::now();
    supercell.buildneighborlist();
    auto thisnlfinish = chrono::high_resolution_clock::now();
    nltime = thisnlfinish-thisnlstart;
    totnltime += nltime;
    neighlistbuilds++;
    
// If necessary, execute a "pushoff" to suppress divergent forces associated with particles' random initial positions.
// After pushoff is complete, give particles random initial velocities corresponding to T = Ttarg
    if (!readresflag) {
        auto pushstart = chrono::high_resolution_clock::now();
        FIRE(1,0);   // Option 1: partial modified-FIRE minimization
//        steepest(1);  // Option 2: steepest-descent minimization
        for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + k%d + d] = 2*sqrt(3)*((1.0*rand())/RAND_MAX - .5)*sqrt(Ttarg/mass[k/d]);
        auto pushfinish = chrono::high_resolution_clock::now();
        pushtime = pushfinish-pushstart;
    }
// Alternatively, if began this run from a restart file, need initialize forces
    if (readresflag) {
        #pragma omp parallel for schedule(guided)
        for (k = 0; k < N; k++) getforce(k);
    }
    
// Start the MD simulation, run for nsteps steps
    cout << "Starting run." << endl;
    
// MAIN LOOP
do {
// Set up MSD calculation to measure diffusivity, etc. on step nmsdstart
    if ((!readresflag && (step == 0 || step == nmsdstart)) || (readresflag && step == nmsdstart)) {
        Dinit = D;
        for (k = 0; k < d*N; k++) Rinit[k] = RVF[(3*d)*(k/d) + k%d];
        for (k = 0; k < d*N; k++) initimg[k] = image[k];
        if (step == nmsdstart) readwrite.writeRinit(step);
    }
    
// Check if neighbor list needs to be rebuilt, rebuild it if it does.
    if (supercell.needsrebuild()) {
        auto thisnlstart = chrono::high_resolution_clock::now();
        supercell.buildneighborlist();
        auto thisnlfinish = chrono::high_resolution_clock::now();
        nltime = thisnlfinish-thisnlstart;
        totnltime += nltime;
        neighlistbuilds++;
        avgneigh = 0;
        for (i = 0; i < N; i++) avgneigh += 2.0*numneigh[i]/N;
        totavgneigh += avgneigh;
    }
    
// Calculate and output thermodynamic and other data
    if (step == startstep) cout << "Step T phi P Z Epair namsd ngp f_ov swapsucc swapatt Runtime" << endl;
    if (!(step%nthermo)) {
        getthermo(true);
        auto currtime = chrono::high_resolution_clock::now();
        totelapsed = currtime-start;
        cout << totelapsed.count() << endl;
    }
    if (vhcflag && step >= nmsdstart && !((step-nmsdstart)%nvhc)) getvanHove();

// If desired, apply barostat once every nbaro steps: need set Ptarg by editing above code
    if (baroflag && step > 0 && !(step%nbaro)) {
        auto thisbarostart = chrono::high_resolution_clock::now();
        Berendsen(Ptarg);
        auto thisbarofinish = chrono::high_resolution_clock::now();
        barotime = thisbarofinish-thisbarostart;
        totbarotime += barotime;
    }
    
// If desired, write coordinates and/or restart files at regular intervals
    if (!(step%ndump)) readwrite.writecoords(step);
    if (step > 0 && !(step%nrestart)) readwrite.writerestart(step);
    
// If desired, energy-minimize once every nmin timesteps
    if (minflag && step > 0 && step-nmsdstart >= 0 && !((step-nmsdstart)%nmin)) {
        if (step == nmsdstart) readwrite.outputmin(-1,0,0,0);
        auto minstart = chrono::high_resolution_clock::now();
        FIRE(pow(10,-25.0),0);   // FIRE minimization
//        steepest(pow(10,-25.0));  // Option 2: steepest-descent minimization
        supercell.buildneighborlist();
        getthermo(true);       // re-output pre-minimization thermo data
        auto currtime = chrono::high_resolution_clock::now();
        totelapsed = currtime-start;
        cout << totelapsed.count() << endl;
        auto minfinish = chrono::high_resolution_clock::now();
        mintime = minfinish - minstart;
        totmintime += mintime;
    }
    
// If desired, perform SWAP once every nswap steps
    if (attfrac && !(step%nswap)) {
        auto thisswapstart = chrono::high_resolution_clock::now();
        lastswapsucc = swapsucc;
        SWAP.swapmove();
        if (swapsucc > lastswapsucc) supercell.buildneighborlist();
        lastswapstep = step;
        auto thisswapfinish = chrono::high_resolution_clock::now();
        swaptime = thisswapfinish-thisswapstart;
        totswaptime += swaptime;
    }
    
// Here perform time integration: integrate forward in time by nbet timesteps
    auto thisverlstart = chrono::high_resolution_clock::now();
    for (i = 0; i < nbet; i++) leapfrog();
    auto thisverlfinish = chrono::high_resolution_clock::now();
    verltime = thisverlfinish-thisverlstart;
    totverltime += verltime;
    step += nbet;
} while (step < nsteps+1);  // End of the MAIN LOOP
} // End of the "master" pragma

// Dethread and wrap up the simulation
   omp_set_num_threads(1);
   readwrite.closefiles();
   if (attfrac) SWAP.stopswap();
   dealloc();
    
// High resolution timing and diagnostics
   auto finish = chrono::high_resolution_clock::now();
   totelapsed = finish-start;
   cout << neighlistbuilds << " neighbor list builds" << endl;
   cout << "Average # of neighbors = " << totavgneigh/neighlistbuilds << endl;
   cout << "Execution took " << totelapsed.count() << " seconds." <<  endl;
   cout << "Pushoff took " << pushtime.count() << " seconds.  Percentage = " << 100*pushtime.count()/totelapsed.count() << endl;
   cout << "Leapfrog timestepping took " << totverltime.count() << " seconds: Percentage = " << 100*totverltime.count()/totelapsed.count() << endl;
   cout << "Neigh-list builds took " << totnltime.count() << " seconds: Percentage = " << 100*totnltime.count()/totelapsed.count() << endl;
   cout << "Minimizations took " << totmintime.count() << " seconds.  Percentage = " << 100*totmintime.count()/totelapsed.count() << endl;
   cout << "Barostatting took " << totbarotime.count() << " seconds: Percentage = " << 100*totbarotime.count()/totelapsed.count() << endl;
   cout << "SWAPping took " << totswaptime.count() << " seconds: Percentage = " << 100*totswaptime.count()/totelapsed.count() << endl;
   cout << "Other processes took " << totelapsed.count() - totnltime.count() - totverltime.count() - pushtime.count() - totmintime.count() - totbarotime.count() - totswaptime.count() << " seconds: Percentage = " << 100*(totelapsed.count() - totnltime.count() - totverltime.count() - pushtime.count() - totmintime.count() - totbarotime.count() - totswaptime.count())/totelapsed.count() << endl;
   for (i = 0; i < 3; i++) cout << endl;
    
   return 0;
}



// Leapfrog integrator
// Includes a Berendsen thermostat for NVT runs
void leapfrog() {
    int i, k;
    double K = 0;  // total kinetic energy of system
    double Ktarg = d*N*Ttarg/2;
    double dampstepT = tau_temp/dt;     // Berendsen thermostat damping time
    double psum[d];     // for momentum-zeroing
#pragma omp parallel
    {
// First half of leapfrog; update v and x, map atoms back into box, reset all forces to zero
    #pragma omp for schedule(static,nchunk) reduction(+:K)
    for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d + d] += dthalf*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
        RVF[(3*d)*(k/d) + k%d] += dt*RVF[(3*d)*(k/d) + k%d + d];
        if (RVF[(3*d)*(k/d) + k%d] > hi) {  RVF[(3*d)*(k/d) + k%d] -= D;   image[k] += 1;  }
        else if (RVF[(3*d)*(k/d) + k%d] < lo) {  RVF[(3*d)*(k/d) + k%d] += D;   image[k] -= 1;  }
        RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
        if (!baroflag && !(step%nbaro)) K += mass[k/d]*RVF[(3*d)*(k/d) + k%d + d]*RVF[(3*d)*(k/d) + k%d + d]/2;
    }
    double lambda;  // Berendsen velocity-scaling factor
    if (!baroflag && !(step%nbaro)) lambda = sqrt(1 + nbaro*(Ktarg/K - 1)/dampstepT);
        
// Update forces
    #pragma omp for schedule(guided)
    for (k = 0; k < N; k++) getforce(k);
        
// Second half of leapfrog: update v
// Apply Berendsen velocity-rescaling factor ONLY if NOT also employing Berendsen baro/thermostat
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d + d] += dthalf*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
        if (!baroflag  && !(step%nbaro)) RVF[(3*d)*(k/d) + k%d + d] *= lambda;
    }
        
// Zero out the total system momentum
// Only really need this for very sensitive measurements of msd (e.g. for glass-transition studies)
// Otherwise can comment it out (since this is also done every time NLs get built)
    #pragma omp for schedule(static,nchunk) reduction(+:psum)
    for (k = 0; k < d*N; k++) psum[k%d] += mass[k/d]*RVF[(3*d)*(k/d) + d + k%d];
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + d + k%d] -= psum[k%d]/(N*mass[k/d]);
    } // end of the parallel region
    
    return;
}



// Calculate force on atom i
void getforce(int i) {
    int j, k, l, u;
    double f[d]; // force on atom i due to atom u
    double fovr;  // this force's magnitude/del
    double delc[d],delc0[d],delsq, del;
    double epsilonij, sigmaij, sigmaijsq; // For polydisp force-mixing
    double facij;
    const double EXPA = exp(alpha);
    
// Calculate force f on atom i by looping over its neighbors (j)
    for (k = 0; k < d; k++) f[k] = 0;
// Defining delc0 here reduces # of sequential accesses of distant parts of memory
    for (k = 0; k < d; k++) delc0[k] = -RVF[3*d*i+k];
// Loop over atom i's neighbors
    for (j = 0; j < numneigh[i]; j++) {
        u = allneigh[i*maxneigh + j];
        sigmaij = (sigma[i]+sigma[u])/2;
        sigmaijsq = sigmaij*sigmaij;
        epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
        delsq = 0;
        #pragma ivdep
        for (k = 0; k < d; k++) {
            delc[k] = delc0[k] + RVF[3*d*u+k];
            if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
            delsq += delc[k]*delc[k];
            if (d > 5) if (delsq > cutsq*sigmaijsq) goto nextneigh;
        }
        if (d > 5 || delsq < cutsq*sigmaijsq) {  // Mention this in our paper
/*  Generalized Lennard-Jones potential
            facij = mypow3(sigmaijsq/delsq);
            fovr = 2*n*epsilonij*facij*(facij-1)/delsq;     */
/* Morse potential */
            del = sqrt(delsq/sigmaijsq);
            facij = EXPA*exp(-alpha*del);
            fovr = (2*alpha*epsilonij/(del*sigmaij))*facij*(facij-1);  // This is |f_ij|/r_ij
// Below works for any radial force
            #pragma ivdep
            for (k = 0; k < d; k++) {
                f[k] = -delc[k]*fovr;
                RVF[3*d*i+2*d+k] += f[k];    RVF[3*d*u+2*d+k] -= f[k];
            }
        }
        nextneigh: ;
    }

    return;
}



// For force & energy calcs. for generalized-LJ:
// This version works and is optimized for arbitrary EVEN n
inline double mypow3(double x) {
    double y = 1.0;
    int j = n/2;  // this edit of mypow2 assumes n is even!!
    int i;
    for (;;) {
        if (j&1) y *= x;
        j >>= 1;
        if (!j) break;
        x *= x;
    }
    return y;
}



// Calculate thermodynamic quantities
void getthermo(bool outputflag) {
    double e, ep, ek, t, p, z, msdincr, namsdincr, m4d;  // per-atom versions of E, EP, EK, etc.
    int i, j, k, u;
    double delc[d], delc0[d], delsq, del;
    double ft, fovr;
    double f[d];
    int totneigh;
    double sigmaij, sigmaijsq, epsilonij; // for polydisp force-mixing
    double facij, shift; // facij = (sigmaij/del)^n
    const double EXPA = exp(alpha);
    for (i = 0; i < N; i++) Z_array[i] = 0;
/* For generalized-LJ
    facij = mypow3(1.0/cutsq);  */
/* For Morse  */
    del = cut;
    facij = EXPA*exp(-alpha*del);
// For both LJ and Morse
    shift = facij*(facij-2);  // shift sets E = 0 at r_ij = cut*sigmaij

// initialize thermo quantities
    EK = 0;     T = 0;      EP = 0;     P = 0;      Z = 0;      Znoratt = 0;
    msd = 0;    namsd = 0;      m4d = 0;    maxdisp = 0;
    f_ov = 0;  // overlap function (Ninarello PRX 2017 Eq. 16)
    rattlers = 0;      floaters = 0; 
    double maxsqdisp = 0;  // maxsqdisp is max (nonaffine displacement)^2 of any atom

#pragma omp parallel
{
    #pragma vector aligned
#pragma omp for schedule(guided) private(j,k,u,ek,t,sigmaij,sigmaijsq,epsilonij,delc,delc0,delsq,del,facij,ep,ft,fovr,f,p) reduction(+:EK,T,Z,EP,P,msd,namsd,m4d,f_ov)
    for (i = 0; i < N; i++) {
        ek = 0;
        for (k = 0; k < d; k++) ek += mass[i]*RVF[3*d*i+d+k]*RVF[3*d*i+d+k]/2;
        t = (2*ek)/d;
        EK += ek/N;       T += t/N;
// Set force on particle i to zero
        for (k = 0; k < d; k++) f[k] = 0;
// declaring delc0[k] here increases efficiency
        for (k = 0; k < d; k++) delc0[k] = -RVF[3*d*i+k];
// Loop over particle i's neighbors
// NOTE INCREMENTS FOR Z, EP, P, HAVE BEEN DOUBLED BECAUSE CODE APPLIES NEWTON'S THIRD LAW
        for (j = 0; j < numneigh[i]; j++) {
            u = allneigh[i*maxneigh + j];
            sigmaij = (sigma[i] + sigma[u])/2.0;
            sigmaijsq = sigmaij*sigmaij;
            epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
            for (k = 0; k < d; k++) delc[k] = delc0[k] + RVF[3*d*u+k];
            for (k = 0; k < d; k++) {
                if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
            }
            delsq = 0;
            for (k = 0; k < d; k++) delsq += delc[k]*delc[k];
            if (delsq <= cutsq*sigmaijsq) {
                Z_array[i] += 1;
                Z_array[u] += 1;
                Z += 2.0/N;
/* For LJ potential
                facij = mypow3(sigmaijsq/delsq);
                fovr = 2*n*epsilonij*facij*(facij-1)/delsq;  */
/* For Morse */
                del = sqrt(delsq/sigmaijsq);
                facij = EXPA*exp(-alpha*del);
                fovr = (2*alpha*epsilonij/(del*sigmaij))*facij*(facij-1);
                // For both Morse and LJ
                ep = epsilonij*(facij*(facij-2) - shift);
                EP += 2*ep/(2*N);  // Should these factors of 2 be elimiinated?
                for (k = 0; k < d; k++) f[k] = -delc[k]*fovr;
                // Here calculate pair contribution to P using virial formula.
                p = delsq*fovr;
                P += p/(d*vol);
            }
        }
// Calculate MSD (total + its nonaffine component)
        double dispthis[d], nadispthis[d];
        double sqdisp = 0;
        double ov_crit = 0;
        for (k = 0; k < d; k++) {
            dispthis[k] = (RVF[3*d*i + k] + image[d*i + k]*D - Rinit[d*i + k] - initimg[d*i + k]*Dinit);
            nadispthis[k] = (RVF[3*d*i + k] + D*image[d*i + k] - (D/Dinit)*Rinit[d*i + k] - Dinit*initimg[d*i + k]);
            namsd += nadispthis[k]*nadispthis[k]/N;
            sqdisp += nadispthis[k]*nadispthis[k];
            ov_crit += nadispthis[k]*nadispthis[k];
        }
        if (sqdisp > maxsqdisp) maxsqdisp = sqdisp;  // maxdisp still needs to be debugged?
        msd += sqdisp/N;
        m4d += sqdisp*sqdisp/N;
// f_ov as defined in Ninarello et al. Physical Review X 2017 (Eq. 16), modified by Adhikari et. al. J. Phys. Chem. B 2021
        if (sqrt(ov_crit) < 0.1*d) f_ov += 1.0/N;
    } // End of loop over the N particles
# pragma omp barrier
} // End of the parallel region

// "Floaters" are atoms with zero contacts.
    for (i = 0; i < N; i++) if (Z_array[i] <= d) {
        rattlers++;
        if (Z_array[i] == 0) floaters++;
        }

// Znoratt =  average coordination number of non-rattlers
    for (i = 0; i < N; i++) if (Z_array[i] > d) Znoratt += Z_array[i]/(1.0*(N - rattlers));

    
// Add the kinetic contribution to pressure and energy
    P += rho*T;
    E = EK + EP;
// For outputting phi during NPT runs
    vol = pow(D,d);
    phi = totpartvol/vol;
// Calculate non-Gaussian param & maximum displacement
    ngp = ((1.0*d)/(d+2))*m4d/(msd*msd) - 1.0;
    if (std::isnan(ngp)) ngp = 0;
    maxdisp = sqrt(maxsqdisp);
    
// Output thermodynamics summary
    if (outputflag) {
        cout << step << " " << T << " " << phi << " " << P << " " << Z << " " << EP << " " << msd << " " << ngp << " " << f_ov << " " << swapsucc << " " << swapatt << " ";
        readwrite.outputthermo(step);
    }
    
    return;
}


// Calculate the van Hove correlation function
void getvanHove() {
    int i, k, bin;
    double vhc[2500], dispthis[d];
    double dispsq, disp;  // displacement of atom i
    for (i = 0; i < 2500; i++) vhc[i] = 0;
    
    #pragma omp parallel for schedule(static,nchunk) private(dispthis,dispsq,k,disp,bin) reduction(+:vhc)
    for (i = 0; i < N; i++) {
        dispsq = 0;
        for (k = 0; k < d; k++) {
            dispthis[k] = (RVF[3*d*i + k] + image[d*i + k]*D - (D/Dinit)*Rinit[d*i + k] - D*initimg[d*i + k]);
            dispsq += dispthis[k]*dispthis[k];
        }
        disp = sqrt(dispsq);
        bin = round(disp/.01);  // note the use of C++'s "round" function
        if (bin < 2500) vhc[bin] += 1.0/N;
    }
    readwrite.outputVHC(vhc,step);
    return;
}


// Berendsen barostat
// This version implicitly assumes the isothermal compressibility kappa = 1
void Berendsen(double Ptarg) {
// Call getthermo() if it wasn't already called this timestep
    if (step%nthermo) getthermo(false);
    rho = N/vol;
    double Pdamp = 18.257; // Berendsen barostat pressure damping time
    double Tdamp = 5.7735; // Berendsen thermostat temperature damping time
    double dampstepP = tau_press/dt;
    double dampstepT = tau_temp/dt;
    double deltaP = P - Ptarg;
    double deltaT = T - Ttarg;
    double lambda, Vexpfac, Lexpfac; // Berensen velocity, volume, and length scaling factors
    int k;
#pragma omp master
{
    Vexpfac = 1 + nbaro*deltaP/dampstepP;
    Lexpfac = pow(Vexpfac,1.0/d);
    lambda = 1 - nbaro*(deltaT/T)/dampstepT;
// Adjust V based on sign and magniture of delta P, # of steps between barostat calls
    vol *= Vexpfac;
    D *= Lexpfac;   // adjust side length of cubic cells
    lo = -D/2;  hi = D/2;
}
// Parallelize the position, velocity updates
    #pragma omp parallel
{
// Scale all positions by Lexpfac and velocities, map back into box if necessary
// Then scale all velocities by lambda/Lexpfac
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d] *= Lexpfac;
        RVF[(3*d)*(k/d) + k%d + d] *= lambda/Lexpfac;
    }
}
// Adjust number of cells if needed
    int ncelltmp = floor(D/(cut*ALU[2] + skin));
    if (ncelltmp != ncellprev) {
        ncell = ncelltmp;
        supercell.setupsupercell();
        supercell.buildneighborlist();
        neighlistbuilds++;
        ncellprev = ncell;
    }
    else if (supercell.needsrebuild()) {
        supercell.buildneighborlist();
        neighlistbuilds++;
    }
    
    return;
}
