 // minimize.cpp -- various methods of energy minimization
#include"omp.h"
#include"cellmanagement.h"
#include"IO.h"
#include<mm_malloc.h>
// Global variables
int iteration;  // Current iteration of minimization routine
int maxiter = 100000;  // End minimizations if they fail to converge after maxiter steps
double *RVFbegin;    // Store pre-minimization coords, velocities, forces
double *Rinitbegin;  // For calculation of particle displacements during minimization
short int *imagebegin; // Store pre-minimization image flags
short int *initimgbegin; // Store pre-min reference-image flags
extern cell supercell;
extern IO readwrite;
extern double dt;
extern double dispcap;
extern double T, E, EP, P, Z, Znoratt;   // Track thermodynamics during minimzation
extern int rattlers, floaters;
extern double min_actual_time;
double dispcapSD;  // separate displacement cap for steepest-descent


// Functions defined in minimize.cpp
void FIRE(double Ethres, double fracDEcap);
void FIREintegrate(double alpha, double f_norm, double v_norm, double dtm, double FdotV);
void onestepdownhill(double f_max);
void steepest(double Ethres);


// Functions defined in other routines
extern void getforce(int i);
extern void getthermo(bool outputflag);


// FIRE minimization routine
// Is a modified version of that described in Guenole et. al, Comp. Mat. Sci (2020)
void FIRE(double Ethres, double fracDEcap) {
    int i, k, steps_since_negative_P, steps_since_under_fracDEcap;
    double dtm = dt;
    double FdotV, alpha, f_norm, f_squared, v_norm, v_squared;
    double f_norm_last = 1000000;
    double forcecap = .00001;
    double EP_last, EP_del;
    bool convflag = false;
    auto startmin = chrono::high_resolution_clock::now();
    chrono::duration<double> totelapsedmin;
    
    cout << "Initializing FIRE minimization" << endl;
    cout << "Iteration T Epair Pressure Z Z_noratt f_norm/(d*N) EP_del/EP Rattlers Floaters Runtime" << endl;
// RVFbegin, Rinitbegin store the state of the MD run at the beginning of minimization
    RVFbegin = (double*) _mm_malloc(3*d*N*sizeof(double),64);
    for (k = 0; k < 3*d*N; k++) RVFbegin[k] = RVF[k];
    Rinitbegin = (double*) _mm_malloc(d*N*sizeof(double),64);
    for (k = 0; k < d*N; k++) Rinitbegin[k] = Rinit[k];
// Also store the initial image, reference-image flags
    imagebegin = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    initimgbegin = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    for (k = 0; k < d*N; k++) {  imagebegin[k] = image[k];  initimgbegin[k] = initimg[k]; }
// Set the temporary Rinit and image flags for tracking particle displacements during minimization
    for (k = 0; k < d*N; k++) { Rinit[k] = RVF[(3*d)*(k/d) + k%d];  image[k] = 0;   initimg[k] = 0; }
    
// Set initial velocities and various counters to zero
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + k%d + d] = 0;
    alpha = .1;
    iteration = 0;
    steps_since_negative_P = 0;
    steps_since_under_fracDEcap = 0;
    min_actual_time = 0;
    
// Execute main loop
    do {
// rebuild NL if necessary
        if (supercell.needsrebuild()) supercell.buildneighborlist();
// Calculate F.v, f_norm, v_norm
        FdotV = 0;      v_squared = 0;      f_squared = 0;
 // Need parallelize this!
        for (k = 0; k < d*N; k++){
            v_squared += RVF[(3*d)*(k/d) + k%d + d]*RVF[(3*d)*(k/d) + k%d + d];
            f_squared += RVF[(3*d)*(k/d) + k%d + 2*d]*RVF[(3*d)*(k/d) + k%d + 2*d];
            FdotV += RVF[(3*d)*(k/d) + k%d + d]*RVF[(3*d)*(k/d) + k%d + 2*d];
        }
        f_norm = sqrt(f_squared);        v_norm = sqrt(v_squared);
// If F.v has been positive for more than 10 consecutive timesteps, increase dtm and decrease alpha
// See Section 3.3 of Guenole 2020
        if (FdotV > 0) {
            steps_since_negative_P++;
            if (steps_since_negative_P > 10) {
                dtm = min(dtm*1.05, max(12-d,5)*dt);
                alpha *= 0.99;
            }
        }
// If F.v is nonpositive, move the positions back a half-step, set all velocities to zero, reset alpha and reduce timestep
        else if (FdotV <= 0) {
            steps_since_negative_P = 0;
            for (k = 0; k < d*N; k++) {
                RVF[(3*d)*(k/d) + k%d] -= 0.5*dtm*RVF[(3*d)*(k/d) + k%d + d];
                if (RVF[(3*d)*(k/d) + k%d] > hi) { RVF[(3*d)*(k/d) + k%d] -= D; image[k] += 1; }
                else if (RVF[(3*d)*(k/d) + k%d] < lo) { RVF[(3*d)*(k/d) + k%d] += D; image[k] -= 1; }
                RVF[(3*d)*(k/d) + k%d + d] = 0;
            }
            alpha = 0.1;
            dtm = max(dtm*.5, dt);  // The FIRE timestep probably shouldn't drop below the MD timestep
        }
// Integrate the FIRE equations of motion
        FIREintegrate(alpha, f_norm, v_norm, dtm, FdotV);
        
// Calculate time in FIRE
        min_actual_time += dtm;
 
// Store previous pair energy and update thermodynamic data
// Then compare current to previous pair energy for convergence check
        EP_last = EP;
        getthermo(false);
        EP_del = EP - EP_last;
// Output thermo data and run-progress summary
        auto currmin = chrono::high_resolution_clock::now();
        totelapsedmin = currmin-startmin;
        if (!(iteration%1000)) {
            cout << iteration << " " << T << " " << EP << " " << P << " " << Z << " " << Znoratt << " ";
            cout << f_norm/(d*N) << " " << EP_del/EP << " " << rattlers << " " << floaters << " " << totelapsedmin.count() << endl;
        }
        readwrite.outputmin(step,iteration,f_norm,EP_del/EP);

// Check for convergence
        if (fabs(EP_del/EP) <= fracDEcap) steps_since_under_fracDEcap++;
        else steps_since_under_fracDEcap = 0;
        if (EP <= Ethres || steps_since_under_fracDEcap > 9) {    convflag = true;     break;     }

// Increment iteration #, break if it exceeds max. value
        iteration++;
        if (iteration > maxiter) {
            cout << "System did not converge. Minimization routine stopped" << endl;
            break;
        }
    } while (!convflag);    // end of mod-FIRE loop
    
// Output final thermo data if it wasn't already output
    auto currmin = chrono::high_resolution_clock::now();
    totelapsedmin = currmin-startmin;
    if (iteration%1000) {
        cout << iteration << " " << T << " " << EP << " " << P << " " << Z << " " << Znoratt << " ";
        cout << f_norm/(d*N) << " " << EP_del/EP << " " << rattlers << " " << floaters << " " << totelapsedmin.count() << endl;
    }

    readwrite.outputmin(step,-1,0,0);

// Restore RVF and image flags to their pre-minimization state and continue the MD simulation
// The if-statement caveat was added so this routine can be used for the initial pushoff
    if (step > 0) {
        for (k = 0; k < 3*d*N; k++) RVF[k] = RVFbegin[k];
        for (k = 0; k < d*N; k++) Rinit[k] = Rinitbegin[k];
        for (k = 0; k < d*N; k++) {     image[k] = imagebegin[k];   initimg[k] = initimgbegin[k];   }
        cout << "Step T phi P Z Epair namsd ngp f_ov swapsucc swapatt Runtime" << endl;
    }
    _mm_free(RVFbegin);         _mm_free(Rinitbegin);
    _mm_free(imagebegin);       _mm_free(initimgbegin);
    
    return;
}



// Semi-implicit Euler integration of the FIRE equations of motion (see Guenole 2020)
void FIREintegrate(double alpha, double f_norm, double v_norm, double dtm, double FdotV) {
    int i, j, k;
#pragma omp parallel
 {
// Update the velocities: cap Δv at 1 along each direction
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
        if (dtm*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d] > 1) RVF[(3*d)*(k/d) + k%d + d] += 1;
        else if (dtm*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d] < -1) RVF[(3*d)*(k/d) + k%d + d] -= 1;
        else RVF[(3*d)*(k/d) + k%d + d] += dtm*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
    }
// Apply the mixing rule if F.v > 0
    if (FdotV > 0) {
        #pragma omp for schedule(static,nchunk)
        for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d + d] = (1 - alpha)*RVF[(3*d)*(k/d) + k%d + d] + alpha*(v_norm/f_norm)*RVF[(3*d)*(k/d) + k%d + 2*d];
        }
    }
     
// Set the velocity of floaters to 0
#pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
        if (Z_array[k/d] == 0) {
            RVF[3*d*(k/d) + k%d + d] = 0;
        }
    }
     
// Update positions with step-size cap = dispcap, map back into box if necessary
   #pragma omp for schedule(static,nchunk)
   for (k = 0; k < d*N; k++) {
        if (RVF[(3*d)*(k/d) + k%d + d] < 0) {
            if (dtm*RVF[(3*d)*(k/d) + k%d + d] > -dispcap) RVF[(3*d)*(k/d) + k%d] += dtm*RVF[(3*d)*(k/d) + k%d + d];
            else RVF[(3*d)*(k/d) + k%d] -= dispcap;
        }
        else if (RVF[(3*d)*(k/d) + k%d + d] > 0) {
            if (dtm*RVF[(3*d)*(k/d) + k%d + d] < dispcap) RVF[(3*d)*(k/d) + k%d] += dtm*RVF[(3*d)*(k/d) + k%d + d];
            else RVF[(3*d)*(k/d) + k%d] += dispcap;
	    }
// Map particles back into box and update image flags if necessary
        if (RVF[(3*d)*(k/d) + k%d] > hi) { RVF[(3*d)*(k/d) + k%d] -= D;  image[k] += 1; }
        else if (RVF[(3*d)*(k/d) + k%d] < lo) { RVF[(3*d)*(k/d) + k%d] += D;    image[k] -= 1; }
    }
// Reset forces to zero and then update them
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
    #pragma omp for schedule(guided)
    for (i = 0; i < N; i++) getforce(i);
 }
    return;
}



// Steepest-descent minimization: like downhill but incorporates an adaptive timestep
// See https://manual.gromacs.org/documentation/2019/reference-manual/algorithms/energy-minimization.html
void steepest(double Ethres) {
    int i, k, iteration;
    double EP_last, EP_del;
    double f_norm, f_squared, f_max, f_this;
    double f_norm_last = 1000000;
    auto startmin = chrono::high_resolution_clock::now();
    chrono::duration<double> totelapsedmin;
    
    cout << "Initializing steepest-descent minimization." << endl;
    cout << "Iteration dispcap Epair Pressure Z Z_noratt f_norm/(d*N) EP_del/EP Rattlers Floaters Runtime" << endl;

// RVFbegin, Rinitbegin store the state of the MD run at the beginning of minimization
    RVFbegin = (double*) _mm_malloc(3*d*N*sizeof(double),64);
    for (k = 0; k < 3*d*N; k++) RVFbegin[k] = RVF[k];
    Rinitbegin = (double*) _mm_malloc(d*N*sizeof(double),64);
    for (k = 0; k < d*N; k++) Rinitbegin[k] = Rinit[k];
// Also store the initial image, reference-image flags
    imagebegin = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    initimgbegin = (short int*) _mm_malloc(d*N*sizeof(short int),16);
    for (k = 0; k < d*N; k++) {  imagebegin[k] = image[k];  initimgbegin[k] = initimg[k]; }
// Set the temporary Rinit and image flags for tracking particle displacements during minimization
    for (k = 0; k < d*N; k++) { Rinit[k] = RVF[(3*d)*(k/d) + k%d];  image[k] = 0;   initimg[k] = 0; }
    
// Execute the main minimization loop
    iteration = 0;  dispcapSD = .01;
    min_actual_time = 0;
    do {
// Calculate f_norm -- currently this is only used for diagnostics
        f_squared = 0;
        #pragma omp parallel for schedule(static,nchunk) private(k) reduction(+:f_squared)
        for (k = 0; k < d*N; k++){
            f_squared += RVF[(3*d)*(k/d) + k%d + 2*d]*RVF[(3*d)*(k/d) + k%d + 2*d];
        }
        f_norm = sqrt(f_squared);
// Calculate f_max
        f_max = 0;
        for (i = 0; i < N; i++) {
            f_this = 0;
            for (k = 0; k < d; k++) f_this += RVF[3*d*i + k + 2*d]*RVF[3*d*i + k + 2*d];
            if (f_this > f_max) f_max = f_this;
        }
// Take one step downhill
        onestepdownhill(f_max);
        if (supercell.needsrebuild()) supercell.buildneighborlist();
        min_actual_time += dispcapSD;
        
// Store previous pair energy and update thermodynamic data
// Then compare current to previous pair energy for convergence check
        EP_last = EP;
        getthermo(false);
        EP_del = EP - EP_last;
// SIMPLEST FORM OF ADAPTIVE TIMESTEPPING
        if (iteration > 0) {
// max. value of dispcap should be skin/2?
// No, this depends on the potential employed: in general should use an energy criterion
            if (EP_del < 0 && dispcapSD < skin/2.05) dispcapSD *= 1.025;
            else if (EP_del > 0) {
// If energy increased during the previous step, take system a step backwards
// this is just the code from onestepdowhill but with a reversed timestep
                #pragma omp for schedule(static,nchunk)
                for (k = 0; k < d*N; k++) {
                    RVF[(3*d)*(k/d) + k%d] -= (RVF[(3*d)*(k/d) + k%d + 2*d]/f_max)*dispcapSD;
                    if (RVF[(3*d)*(k/d) + k%d] > hi) {  RVF[(3*d)*(k/d) + k%d] -= D;   image[k] += 1;  }
                    else if (RVF[(3*d)*(k/d) + k%d] < lo) {  RVF[(3*d)*(k/d) + k%d] += D;   image[k] -= 1;  }
                }
// Reduce the maximum step size: need add a minimum to prevent runaway dispcapSD --> 0  ?
// Could add a dispcap-based convergence criterion
                dispcapSD *= .5;
// Recalc. EP and EP_del here to avoid spurious output below
                getthermo(false);
                EP_del = EP - EP_last;
            }
        }
// Output thermo data and run-progress summary
        auto currmin = chrono::high_resolution_clock::now();
        totelapsedmin = currmin-startmin;
/* FIRE-STYLE OUTPUT */
        if (!(iteration%1000)) {
             cout << iteration << " " << dispcapSD << " " << EP << " " << P << " " << Z << " " << Znoratt << " ";
             cout << f_norm/(d*N) << " " << EP_del/EP << " " << rattlers << " " << floaters << " " << totelapsedmin.count() << endl;
         }
        readwrite.outputmin(step,iteration,f_norm,EP_del/EP);
        iteration++;
    } while (EP > Ethres && iteration < maxiter+1);
    if (iteration > maxiter) {
        cout << "System did not converge. Minimization routine stopped" << endl;
    }

    readwrite.outputmin(step,-1,0,0);
    
// Restore RVF and image flags to their pre-minimization state and continue the MD simulation
// The if-statement caveat was added so this routine can be used for the initial pushoff
    if (step > 0) {
        for (k = 0; k < 3*d*N; k++) RVF[k] = RVFbegin[k];
        for (k = 0; k < d*N; k++) Rinit[k] = Rinitbegin[k];
        for (k = 0; k < d*N; k++) {     image[k] = imagebegin[k];   initimg[k] = initimgbegin[k];   }
        cout << "Step T phi P Z Epair namsd ngp f_ov swapsucc swapatt Runtime" << endl;
    }
    _mm_free(RVFbegin);         _mm_free(Rinitbegin);
    _mm_free(imagebegin);       _mm_free(initimgbegin);

// reset dispcap to initial value
    dispcapSD = .01;
    
    return;
}



// Straight-downhill minimization step
void onestepdownhill(double f_max) {
    int i, j, k;
#pragma omp parallel
 {
// Reset all forces to zero
    #pragma omp for
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
// Update forces
    #pragma omp for schedule(guided)
    for (i = 0; i < N; i++) getforce(i);
// Update positions with step-size cap = dispcapSD, map back into cell if necessary
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
// The textbook method: see https://manual.gromacs.org/documentation/2019/reference-manual/algorithms/energy-minimization.html
        RVF[(3*d)*(k/d) + k%d] += (RVF[(3*d)*(k/d) + k%d + 2*d]/f_max)*dispcapSD;
        if (RVF[(3*d)*(k/d) + k%d] > hi) {  RVF[(3*d)*(k/d) + k%d] -= D;   image[k] += 1;  }
        else if (RVF[(3*d)*(k/d) + k%d] < lo) {  RVF[(3*d)*(k/d) + k%d] += D;   image[k] -= 1;  }
    }
 }
    return;
}
